"""Widget emulating a segment display."""
from typing import Optional, List, Dict, Any
import math

from kivy.clock import Clock
from kivy.graphics.vertex_instructions import Mesh, Ellipse, Rectangle
from kivy.properties import NumericProperty, BooleanProperty, ListProperty, StringProperty, OptionProperty
from kivy.graphics import Color, Rotate, Scale
from kivy.utils import get_color_from_hex

from mpf.core.segment_mappings import FOURTEEN_SEGMENTS, SEVEN_SEGMENTS, EIGHT_SEGMENTS
from mpf.core.utility_functions import Util

from mpfmc.uix.widget import Widget

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc             # pylint: disable-msg=cyclic-import,unused-import

# Constants for punctuation segments (dot/period and comma)
OFF = 0


# pylint: disable-msg=too-many-instance-attributes
class SegmentDisplayEmulator(Widget):

    """Widget emulating a segment display."""

    widget_type_name = 'SegmentDisplayEmulator'
    merge_settings = ('height', 'width')
    animation_properties = ('x', 'y', 'scale', 'width', 'height', 'opacity', 'rotation', 'segment_on_color')

    display_instances = []

    def __init__(self, mc: "MpfMc", config: dict, key: Optional[str] = None, **kwargs) -> None:
        super().__init__(mc=mc, config=config, key=key)

        # Initialize the character segment map with the default 14-segment mappings from MPF
        if self.display_type == "7seg":
            self._segment_map = {k: self.get_seven_segment_character_encoding(v) for k, v in SEVEN_SEGMENTS.items()}
            self._segment_count = 7
            self._dot_segment_index = 0
            self._comma_segment_index = 7
        elif self.display_type == "8seg":
            self._segment_map = {k: self.get_eight_segment_character_encoding(v) for k, v in EIGHT_SEGMENTS.items()}
            self._segment_count = 8
            self._dot_segment_index = 8
            self._comma_segment_index = 9
        elif self.display_type == "14seg":
            self._segment_map = {k: self.get_fourteen_segment_character_encoding(v) for k, v in
                                 FOURTEEN_SEGMENTS.items()}
            self._segment_count = 14
            self._dot_segment_index = 14
            self._comma_segment_index = 15

        # Override default character segment mappings with specific settings
        if "character_map" in self.config and self.config["character_map"] is not None:
            for k, v in self.config["character_map"].items():
                self._segment_map[k] = v

        # The initial value for segment_on_color is typically set using the config file which creates
        # a list of lists. A flat one-dimensional list is required so flatten the initial list.
        self.segment_on_color = list(Util.flatten_list(self.segment_on_color))

        # Initialize the encoded display characters
        self._encoded_characters = [OFF] * self.character_count

        self._flash_clock_event = None
        self._apply_flash_mask = False
        self._flash_character_mask = [0xFF] * self.character_count
        self._segment_colors = None
        self._segment_mesh_objects = None
        self._update_event_handler_key = None

        # Bind to all properties that when changed need to force
        # the widget to be redrawn
        self.bind(text=self._update_text,
                  flash_mode=self._set_flash_mode,
                  pos=self._draw_widget,
                  size=self._recalculate,
                  background_color=self._draw_widget,
                  segment_off_color=self._draw_widget,
                  segment_on_color=self._set_character_colors,
                  segment_width=self._recalculate,
                  segment_interval=self._recalculate,
                  bevel_width=self._recalculate,
                  side_bevel_enabled=self._recalculate,
                  character_slant_angle=self._recalculate,
                  character_spacing=self._recalculate)

        self._calculate_segment_points()
        self._set_character_colors()
        self._update_text()
        self._set_flash_mode()

        self._update_event_handler_key = self.mc.events.add_handler(
            "update_segment_display",
            self.on_update_segment_display)

    def prepare_for_removal(self) -> None:
        """Prepare the widget to be removed."""
        if self._update_event_handler_key:
            self.mc.events.remove_handler_by_key(self._update_event_handler_key)
            self._update_event_handler_key = None

        # important to call base class to ensure the widget gets removed
        super().prepare_for_removal()

    @staticmethod
    def get_seven_segment_character_encoding(segments: SEVEN_SEGMENTS) -> int:
        """Returns segment value in order used in the segment display widget."""
        # Note: the l and n segments appear to be swapped in the encodings in the FOURTEEN_SEGMENTS dict
        return int(
            (segments.g << 6) | (segments.f << 5) | (segments.e << 4) |
            (segments.d << 3) | (segments.c << 2) | (segments.b << 1) | segments.a)

    @staticmethod
    def get_eight_segment_character_encoding(segments: EIGHT_SEGMENTS) -> int:
        """Returns segment value in order used in the segment display widget."""
        # Note: the l and n segments appear to be swapped in the encodings in the FOURTEEN_SEGMENTS dict
        return int(
            (segments.h << 7) | (segments.g << 6) | (segments.f << 5) | (segments.e << 4) |
            (segments.d << 3) | (segments.c << 2) | (segments.b << 1) | segments.a)

    @staticmethod
    def get_fourteen_segment_character_encoding(segments: FOURTEEN_SEGMENTS) -> int:
        """Returns segment value in order used in the segment display widget."""
        # Note: the l and n segments appear to be swapped in the encodings in the FOURTEEN_SEGMENTS dict
        return int(
            (segments.dp << 14) | (segments.l << 13) | (segments.m << 12) |
            (segments.n << 11) | (segments.k << 10) | (segments.j << 9) | segments.h << 8 |
            (segments.g2 << 7) | (segments.g1 << 6) | (segments.f << 5) | (segments.e << 4) |
            (segments.d << 3) | (segments.c << 2) | (segments.b << 1) | segments.a)

    @staticmethod
    def _flip_horizontal(points, width):
        """Flips the supplied points over the horizontal symmetry line"""
        flipped_points = []

        for index in range(0, len(points), 2):
            flipped_points.append(width - points[index])
            flipped_points.append(points[index + 1])

        return flipped_points

    @staticmethod
    def _flip_vertical(points, height):
        """Flips the supplied points over the vertical symmetry line"""
        flipped_points = []

        for index in range(0, len(points), 2):
            flipped_points.append(points[index])
            flipped_points.append(height - points[index + 1])

        return flipped_points

    @staticmethod
    def _apply_character_slant_to_points(points, slant_slope):
        """Applies slant to x-coordinates of segment points"""
        modified_points = []

        for index in range(0, len(points), 2):
            modified_points.append(points[index] + slant_slope * points[index + 1])
            modified_points.append(points[index + 1])

        return modified_points

    def _recalculate(self, *args):
        """Recalculate the segments and redraw the display widget."""
        del args
        self._calculate_segment_points()
        self._draw_widget()

    def _calculate_segment_points(self):
        """Calculate the points of all the display segments to be drawn."""

        # The coordinate calculations in this module are based on an open source sixteen segment display project:
        # https://github.com/Enderer/sixteensegment

        # Calculate the size dimensions of a single character/element in the display
        self.char_height = self.height - (self.padding * 2)

        # Determine if characters are slanted
        if self.character_slant_angle > 0.0:
            # Use some basic triangle trigonometry to determine the distance the letters are slanted at the top
            slant_angle_radians = self.character_slant_angle * math.pi / 180.0
            slant_opposite_angle_radians = (90 - self.character_slant_angle) * math.pi / 180.0
            slant_horizontal_distance = self.char_height * math.sin(slant_angle_radians) / math.sin(
                slant_opposite_angle_radians)

            # Calculate the slope of the slanted vertical line segments
            slant_slope = slant_horizontal_distance / self.char_height

            # Padding must be greater than or equal to slant_horizontal_distance
            if self.character_spacing < slant_horizontal_distance:
                self.character_spacing = slant_horizontal_distance
        else:
            slant_slope = 0

        # Calculate the width of each character (from corner to corner, not including spacing and padding)
        self.char_width = (self.width - self.character_spacing * (self.character_count - 1) - (
            self.padding * 2)) / self.character_count

        segment_width = self.segment_width * self.char_width
        segment_interval = self.segment_interval * self.char_width
        bevel_width = self.bevel_width * self.char_width

        # Base positions of points without bevel and interval
        x = [self.char_width / 2 - segment_width / 2,
             self.char_width / 2,
             self.char_width / 2 + segment_width / 2,
             self.char_width - segment_width,
             self.char_width - segment_width / 2,
             self.char_width]

        y = [0,
             segment_width / 2,
             segment_width,
             self.char_height / 2 - segment_width / 2,
             self.char_height / 2,
             self.char_height / 2 + segment_width / 2]

        if self.display_type == "14seg":
            segment_points = self._calculate_fourteen_segment_points(x, y, segment_width, segment_interval, bevel_width)
        elif self.display_type == "7seg":
            segment_points = self._calculate_seven_segment_points(x, y, segment_width, segment_interval, bevel_width)
        elif self.display_type == "8seg":
            segment_points = self._calculate_eight_segment_points(x, y, segment_width, segment_interval, bevel_width)

        # Sort the segment dictionary by segment name (key)
        segment_points = dict(sorted(segment_points.items()))

        if self.dot_enabled:
            # Store center point of dot/period circle (diameter will be segment_width) (x, y, radius)
            self._dot_points = [self.char_width + segment_width / 2, 0, segment_width]

        if self.comma_enabled:
            #  (comma should fit in padding on right side of character/element)
            self._comma_points = [self.char_width + 0.5 * segment_width, -0.5 * segment_interval,
                                  self.char_width + 0.75 * segment_width, -1.5 * segment_interval,
                                  self.char_width + segment_width, -2.5 * segment_interval,
                                  self.char_width + 1.25 * segment_width, -1.5 * segment_interval,
                                  self.char_width + 1.5 * segment_width, -0.5 * segment_interval,
                                  self.char_width + 1.4 * segment_width, -0.75 * segment_width,
                                  self.char_width + 0.5 * segment_width, -2 * segment_width,
                                  self.char_width + 0.8 * segment_width, -0.5 * segment_width]

        # Determine if characters are slanted (if so, apply slant translation)
        if self.character_slant_angle > 0.0:
            segment_points.update(
                {k: self._apply_character_slant_to_points(v, slant_slope) for k, v in segment_points.items()})
            if self.comma_enabled:
                self._comma_points = self._apply_character_slant_to_points(self._comma_points, slant_slope)

        self._segment_points = list(segment_points.values())

    # pylint: disable-msg=too-many-arguments
    def _calculate_seven_segment_points(self, x: List[float], y: List[float],
                                        segment_width: float, segment_interval: float,
                                        bevel_width: float) -> Dict[str, float]:
        """Calculate the vertices for all segments in a seven-segment display."""

        side_bevel_multiplier = 1 if self.side_bevel_enabled else 0

        sqrt2 = math.sqrt(2)
        sqrt3 = math.sqrt(3)

        # Create dictionary of segment points keyed by segment name/letter
        segment_points = {"d": [bevel_width * 2 + segment_interval / sqrt2, y[0],
                                x[5] - (bevel_width * 2 + segment_interval / sqrt2), y[0],
                                x[5] - (bevel_width + segment_interval / sqrt2), y[1],
                                x[5] - (bevel_width * 2 + segment_interval / sqrt2), y[2],
                                bevel_width * 2 + segment_interval / sqrt2, y[2],
                                bevel_width + segment_interval / sqrt2, y[1]],
                          "g": [segment_width + segment_interval / 2 * sqrt3, y[3],
                                x[3] - segment_interval / 2 * sqrt3, y[3],
                                x[4] - segment_interval / 2 * sqrt3, y[4],
                                x[3] - segment_interval / 2 * sqrt3, y[5],
                                segment_width / 2 + segment_interval / 2 * sqrt3, y[5],
                                segment_width + segment_interval / 2 * sqrt3, y[4]],
                          "c": [x[5], y[0] + bevel_width * 2 + segment_interval / sqrt2,
                                x[5], y[4] - segment_interval / 2 - segment_width / 2 * side_bevel_multiplier,
                                x[4], y[4] - segment_interval / 2,
                                x[3], y[3] - segment_interval / 2,
                                x[3], y[2] + segment_interval / sqrt2,
                                x[5] - bevel_width, y[0] + bevel_width + segment_interval / sqrt2]}

        # Create the rest of the segments by flipping/mirroring existing points (either horizontally or vertically)
        segment_points["a"] = self._flip_vertical(segment_points["d"], self.char_height)
        segment_points["b"] = self._flip_vertical(segment_points["c"], self.char_height)
        segment_points["e"] = self._flip_horizontal(segment_points["c"], self.char_width)
        segment_points["f"] = self._flip_horizontal(segment_points["b"], self.char_width)

        return segment_points

    # pylint: disable-msg=too-many-arguments
    def _calculate_eight_segment_points(self, x: List[float], y: List[float],
                                        segment_width: float, segment_interval: float,
                                        bevel_width: float) -> Dict[str, float]:
        """Calculate the vertices for all segments in a fourteen-segment display."""

        side_bevel_multiplier = 1 if self.side_bevel_enabled else 0

        sqrt2 = math.sqrt(2)
        sqrt3 = math.sqrt(3)

        # Create dictionary of segment points keyed by segment name/letter
        segment_points = {"d": [bevel_width * 2 + segment_interval / sqrt2, y[0],
                                x[5] - (bevel_width * 2 + segment_interval / sqrt2), y[0],
                                x[5] - (bevel_width + segment_interval / sqrt2), y[1],
                                x[5] - (bevel_width * 2 + segment_interval / sqrt2), y[2],
                                bevel_width * 2 + segment_interval / sqrt2, y[2],
                                bevel_width + segment_interval / sqrt2, y[1]],
                          "g": [x[3] - segment_interval / 2 * sqrt3, y[3],
                                x[4] - segment_interval / 2 * sqrt3, y[4],
                                x[3] - segment_interval / 2 * sqrt3, y[5],
                                bevel_width * 2 + segment_interval / sqrt2, y[5],
                                bevel_width + segment_interval / sqrt2, y[4],
                                bevel_width * 2 + segment_interval / sqrt2, y[3]],
                          "c": [x[5], y[0] + bevel_width * 2 + segment_interval / sqrt2,
                                x[5], y[4] - segment_interval / 2 - segment_width / 2 * side_bevel_multiplier,
                                x[4], y[4] - segment_interval / 2,
                                x[3], y[3] - segment_interval / 2,
                                x[3], y[2] + segment_interval / sqrt2,
                                x[5] - bevel_width, y[0] + bevel_width + segment_interval / sqrt2],
                          "h": [x[0], y[2] + segment_interval,
                                x[2], y[2] + segment_interval,
                                x[2], self.char_height - y[2] - segment_interval,
                                x[0], self.char_height - y[2] - segment_interval]}

        # Create the rest of the segments by flipping/mirroring existing points (either horizontally or vertically)
        segment_points["a"] = self._flip_vertical(segment_points["d"], self.char_height)
        segment_points["b"] = self._flip_vertical(segment_points["c"], self.char_height)
        segment_points["e"] = self._flip_horizontal(segment_points["c"], self.char_width)
        segment_points["f"] = self._flip_horizontal(segment_points["b"], self.char_width)

        return segment_points

    # pylint: disable-msg=too-many-arguments
    def _calculate_fourteen_segment_points(self, x: List[float], y: List[float],
                                           segment_width: float, segment_interval: float,
                                           bevel_width: float) -> Dict[str, float]:
        """Calculate the vertices for all segments in a fourteen-segment display."""

        side_bevel_multiplier = 1 if self.side_bevel_enabled else 0

        segment_factor = segment_width * 0.8
        diagonal_slope = self.char_height / self.char_width
        sqrt2 = math.sqrt(2)
        sqrt3 = math.sqrt(3)

        # Create dictionary of segment points keyed by segment name/letter
        segment_points = {"d": [bevel_width * 2 + segment_interval / sqrt2, y[0],
                                x[5] - (bevel_width * 2 + segment_interval / sqrt2), y[0],
                                x[5] - (bevel_width + segment_interval / sqrt2), y[1],
                                x[5] - (bevel_width * 2 + segment_interval / sqrt2), y[2],
                                bevel_width * 2 + segment_interval / sqrt2, y[2],
                                bevel_width + segment_interval / sqrt2, y[1]],
                          "g2": [x[1] + segment_interval / 2, y[3],
                                 x[3] - segment_interval / 2 * sqrt3, y[3],
                                 x[4] - segment_interval / 2 * sqrt3, y[4],
                                 x[3] - segment_interval / 2 * sqrt3, y[5],
                                 x[1] + segment_interval / 2, y[5]],
                          "c": [x[5], y[0] + bevel_width * 2 + segment_interval / sqrt2,
                                x[5], y[4] - segment_interval / 2 - segment_width / 2 * side_bevel_multiplier,
                                x[4], y[4] - segment_interval / 2,
                                x[3], y[3] - segment_interval / 2,
                                x[3], y[2] + segment_interval / sqrt2,
                                x[5] - bevel_width, y[0] + bevel_width + segment_interval / sqrt2],
                          "m": [x[2], y[2] + segment_interval,
                                x[2], y[3] - segment_interval,
                                x[0], y[3] - segment_interval,
                                x[0], y[2] + segment_interval],
                          "l": [(segment_width + segment_factor) / diagonal_slope + segment_interval,
                                y[2] + segment_interval,
                                x[0] - segment_interval, x[0] * diagonal_slope - segment_factor - segment_interval,
                                x[0] - segment_interval, y[3] - segment_interval,
                                (y[3] - segment_interval) / diagonal_slope - segment_interval, y[3] - segment_interval,
                                segment_width + segment_interval,
                                y[2] * diagonal_slope + segment_factor + segment_interval,
                                segment_width + segment_interval, y[2] + segment_interval]}

        # Create the rest of the segments by flipping/mirroring existing points (either horizontally or vertically)
        segment_points["a"] = self._flip_vertical(segment_points["d"], self.char_height)
        segment_points["b"] = self._flip_vertical(segment_points["c"], self.char_height)
        segment_points["e"] = self._flip_horizontal(segment_points["c"], self.char_width)
        segment_points["f"] = self._flip_horizontal(segment_points["b"], self.char_width)
        segment_points["g1"] = self._flip_horizontal(segment_points["g2"], self.char_width)
        segment_points["h"] = self._flip_vertical(segment_points["l"], self.char_height)
        segment_points["j"] = self._flip_vertical(segment_points["m"], self.char_height)
        segment_points["k"] = self._flip_horizontal(segment_points["h"], self.char_width)
        segment_points["n"] = self._flip_horizontal(segment_points["l"], self.char_width)

        return segment_points

    def _draw_widget(self, *args) -> None:
        """Establish the drawing instructions for the widget."""
        del args

        if self.canvas is None:
            return

        self.canvas.clear()
        self._segment_colors = []
        self._segment_mesh_objects = []

        # Get the list of encoded characters (apply flash mask if applicable)
        encoded_characters = [
            self._encoded_characters[index] & self._flash_character_mask[index] if self._apply_flash_mask else
            self._encoded_characters[index] for index in range(len(self._encoded_characters))]

        with self.canvas:
            Color(*self.background_color)
            Scale(self.scale, origin=self.center)
            Rotate(angle=self.rotation, origin=self.center)

            # Fill background
            Rectangle(pos=self.pos, size=self.size)

            x_offset = self.x + self.padding
            y_offset = self.y + self.padding

            for index, encoded_char in enumerate(encoded_characters):

                colors = [None] * (self._segment_count + 2)
                mesh_objects = [None] * self._segment_count
                for segment in range(self._segment_count):
                    colors[segment] = self._create_segment_color(index, segment, encoded_char)
                    mesh_objects[segment] = self._create_segment_mesh_object(segment, x_offset, y_offset)

                self._segment_colors.append(colors)
                self._segment_mesh_objects.append(mesh_objects)
                if self.dot_enabled:
                    colors[self._dot_segment_index] = self._create_segment_color(index, self._dot_segment_index,
                                                                                 encoded_char)
                    Ellipse(pos=(self._dot_points[0] + x_offset, self._dot_points[1] + y_offset),
                            size=(self._dot_points[2], self._dot_points[2]))

                if self.comma_enabled:
                    colors[self._comma_segment_index] = self._create_segment_color(index, self._comma_segment_index,
                                                                                   encoded_char)
                    Mesh(vertices=[self._comma_points[0] + x_offset, self._comma_points[1] + y_offset, 0, 0,
                                   self._comma_points[2] + x_offset, self._comma_points[3] + y_offset, 0, 0,
                                   self._comma_points[4] + x_offset, self._comma_points[5] + y_offset, 0, 0,
                                   self._comma_points[6] + x_offset, self._comma_points[7] + y_offset, 0, 0,
                                   self._comma_points[8] + x_offset, self._comma_points[9] + y_offset, 0, 0,
                                   self._comma_points[10] + x_offset, self._comma_points[11] + y_offset, 0, 0,
                                   self._comma_points[12] + x_offset, self._comma_points[13] + y_offset, 0, 0,
                                   self._comma_points[14] + x_offset, self._comma_points[15] + y_offset, 0, 0],
                         indices=[0, 1, 2, 3, 4, 5, 6, 7],
                         mode="triangle_fan")

                x_offset += self.char_width + self.character_spacing

    def on_update_segment_display(self, segment_display_name: Any, **kwargs):
        """Event handler method to update the segment display."""
        if segment_display_name == self.config['name']:
            if 'text' in kwargs:
                self.text = kwargs['text']
            if 'colors' in kwargs and kwargs['colors']:
                colors = []
                for color in kwargs['colors']:
                    colors.extend(get_color_from_hex(color))
                self.segment_on_color = colors
            if 'flashing' in kwargs:
                if kwargs['flashing'] == "False":
                    self.flash_mode = "off"
                elif kwargs['flashing'] == "True":
                    self.flash_mode = "all"
                elif kwargs['flashing'] == "match":
                    self.flash_mode = "match"
                elif kwargs['flashing'] == "mask":
                    self.flash_mask = kwargs.get('flash_mask', "")
                    self.flash_mode = "mask"
                else:
                    self.flash_mode = "off"

    def _update_text(self, *args):
        """Process the new text value to prepare it for display"""
        del args
        self._encoded_characters = self.encode_characters(self.text, self.character_count, self._segment_map,
                                                          self.dot_enabled, 1 << self._dot_segment_index,
                                                          self.comma_enabled, 1 << self._comma_segment_index)
        self._draw_widget()

    def _create_segment_color(self, index: int, segment: int, char_code: int):
        """Creates a Color vertex instruction for the specified segment number"""
        if (1 << segment) & char_code:
            return Color(self.segment_on_color[index * 4], self.segment_on_color[index * 4 + 1],
                         self.segment_on_color[index * 4 + 2], self.segment_on_color[index * 4 + 3])
        return Color(self.segment_off_color[0], self.segment_off_color[1], self.segment_off_color[2],
                     self.segment_off_color[3])

    def _create_segment_mesh_object(self, segment, x_offset, y_offset):
        """Creates a Mesh vertex instruction for the specified segment number"""
        points = self._segment_points[segment]
        vertices = []
        indices = []
        for index in range(0, len(points), 2):
            vertices.extend([points[index] + x_offset, points[index + 1] + y_offset, 0, 0])
            indices.append(int(index / 2))

        return Mesh(vertices=vertices, indices=indices, mode="triangle_fan")

    def _create_segment_dot_object(self, x_offset, y_offset):
        pass

    def _update_segment_colors(self):
        """Update the colors for every segment (on or off based on actual characters)."""
        for char_index in range(self.character_count):
            char_code = self._encoded_characters[char_index]
            on_color = self._character_colors[char_index * 4:char_index * 4 + 4]
            for segment in range(16):
                if (1 << segment) & char_code:
                    self._segment_colors[char_index][segment].rgba = on_color
                else:
                    self._segment_colors[char_index][segment].rgba = self.segment_off_color

    # pylint: disable-msg=too-many-arguments
    @staticmethod
    def encode_characters(text: str, character_count: int, segment_map: Dict[int, int],
                          dot_enabled: bool, dot_segment_mask: int,
                          comma_enabled: bool, comma_segment_mask: int) -> List[int]:
        """Encode the text characters to prepare for display."""
        text_position = 0
        encoded_characters = []
        while text_position < len(text):
            char = text[text_position]
            text_position += 1
            encoded_char = segment_map.get(ord(char), 0x00)
            if dot_enabled or comma_enabled:
                # embed dots is enabled and dot is inactive
                try:
                    next_char = text[text_position]
                except IndexError:
                    next_char = " "
                if dot_enabled and next_char == ".":
                    # next char is a dot -> turn dot on
                    encoded_char |= dot_segment_mask
                    text_position += 1
                elif comma_enabled and next_char == ",":
                    # next char is a dot -> turn dot on
                    encoded_char |= comma_segment_mask
                    text_position += 1

            encoded_characters.append(encoded_char)

        # remove leading segments if mapping is too long
        if character_count < len(encoded_characters):
            encoded_characters = encoded_characters[-character_count:]

        while character_count > len(encoded_characters):
            # prepend spaces to pad mapping
            encoded_characters.insert(0, OFF)

        return encoded_characters

    def _set_character_colors(self, *args):
        """Set the segment on/active colors for each character in the display."""
        del args
        self._character_colors = self.segment_on_color
        color_count = int(len(self._character_colors) / 4)

        # fill the rest of the character colors with the last color
        if color_count < self.character_count:
            last_color = self._character_colors[-4:]
            self._character_colors.extend(last_color * (self.character_count - color_count))

        self._draw_widget()

    def _set_flash_mode(self, *args):
        """Set the current flash mode."""
        del args
        if self.flash_mode == "off":
            self._flash_character_mask = [0xFF] * self.character_count
            self._stop_flash_timer()
        elif self.flash_mode == "all":
            self._flash_character_mask = [0x00] * self.character_count
            self._start_flash_timer()
        elif self.flash_mode == "match":
            self._flash_character_mask = [0xFF] * (self.character_count - 2)
            self._flash_character_mask.extend([0x00] * 2)
            self._start_flash_timer()
        elif self.flash_mode == "mask":
            mask = self.flash_mask.rjust(self.character_count, ' ')
            mask = mask[-self.character_count:]
            self._flash_character_mask = [0x00 if char == "F" else 0xFF for char in mask]
            self._start_flash_timer()

    def _start_flash_timer(self):
        """Start the timer to control flashing."""
        self._apply_flash_mask = True
        if self._flash_clock_event:
            self._flash_clock_event.cancel()
        self._flash_clock_event = Clock.schedule_interval(self._flash_clock_callback, 1 / (self.flash_frequency * 2))
        self._draw_widget()

    def _stop_flash_timer(self):
        """Stop the timer that controls flashing."""
        self._apply_flash_mask = False
        if self._flash_clock_event:
            self._flash_clock_event.cancel()
            self._flash_clock_event = None
        self._draw_widget()

    def _flash_clock_callback(self, dt, *args):
        """Callback method for the clock."""
        del dt
        del args
        self._apply_flash_mask = not self._apply_flash_mask
        self._draw_widget()

    #
    # Properties
    #
    rotation = NumericProperty(0)
    '''Rotation angle value of the widget.

    :attr:`rotation` is an :class:`~kivy.properties.NumericProperty` and defaults to
    0.
    '''

    scale = NumericProperty(1.0)
    '''Scale value of the widget.

    :attr:`scale` is an :class:`~kivy.properties.NumericProperty` and defaults to
    1.0.
    '''

    character_count = NumericProperty(1)
    '''The number of display characters in the segment display.

    :attr:`character_count` is an :class:`~kivy.properties.NumericProperty` and defaults to 1.
    '''

    display_type = OptionProperty("14seg", options=["7seg", "8seg", "14seg"])
    '''The type of display (7 segment, 8 segment, 14 segment).

    :attr:`display_type` is an :class:`~kivy.properties.OptionProperty` and defaults to `14SEG`.
    '''

    character_slant_angle = NumericProperty(0)
    '''The angle at which the characters are slanted (degrees from vertical)

    :attr:`character_slant_angle` is an :class:`~kivy.properties.NumericProperty` and defaults to 0.
    '''

    character_spacing = NumericProperty(10)
    '''The space between each character/element.

    :attr:`character_spacing` is an :class:`~kivy.properties.NumericProperty` and defaults to 10.
    '''

    padding = NumericProperty(10)
    '''The padding around the display.

    :attr:`padding` is an :class:`~kivy.properties.NumericProperty` and defaults to 10.
    '''

    segment_width = NumericProperty(0.16)
    '''Width of each segment (as a percentage of character/element width).

    :attr:`segment_width` is an :class:`~kivy.properties.NumericProperty` and defaults to 0.16.
    '''

    segment_interval = NumericProperty(0.05)
    '''Spacing between segments (as a percentage of character/element width).

    :attr:`segment_interval` is an :class:`~kivy.properties.NumericProperty` and defaults to 0.05.
    '''

    bevel_width = NumericProperty(0.06)
    '''Size of bevels (as a percentage of character/element width).

    :attr:`bevel_width` is an :class:`~kivy.properties.NumericProperty` and defaults to 0.06.
    '''

    side_bevel_enabled = BooleanProperty(True)
    '''Determines if the sides should be beveled

    :attr:`side_bevel_enabled` is an :class:`kivy.properties.BooleanProperty` and defaults to True.
    '''

    segment_off_color = ListProperty([0.588, 0.596, 0.584, 1.0])
    '''The color of a segment that is off, in the (r, g, b, a) format.

    :attr:`segment_off_color` is a :class:`~kivy.properties.ListProperty` and
    defaults to [0.588, 0.596, 0.584, 1.0].
    '''

    segment_on_color = ListProperty([0.867, 0.510, 0.090, 1.0])
    '''The color of a segment that is on, in the (r, g, b, a) format. When there is only one color in
    the list, all characters will use that color. The list can be extended to set a color for each
    character in the display. If there are fewer colors than characters, the last color in the list
    will be extended to the remaining characters.

    :attr:`segment_on_color` is a :class:`~kivy.properties.ListProperty` and
    defaults to [0.867, 0.510, 0.090, 1.0].
    '''

    background_color = ListProperty([0, 0, 0, 1.0])
    '''The background color of the display widget, in the (r, g, b, a) format.

    :attr:`background_color` is a :class:`~kivy.properties.ListProperty` and
    defaults to [0, 0, 0, 1.0].
    '''

    dot_enabled = BooleanProperty(False)
    '''Determines if an integrated dot/period should be displayed

    :attr:`dot_enabled` is an :class:`kivy.properties.BooleanProperty` and defaults to False.
    '''

    comma_enabled = BooleanProperty(False)
    '''Determines if an integrated comma should be displayed

    :attr:`comma_enabled` is an :class:`kivy.properties.BooleanProperty` and defaults to False.
    '''

    text = StringProperty("")
    '''Contains the text string to display

    :attr:`text` is an :class:`kivy.properties.StringProperty` and defaults to "".
    '''

    flash_mode = OptionProperty("off", options=["off", "all", "match", "mask"])
    '''The current disiplay flash mode. Defaults to "off".

    :attr:`flash_mode` is an :class:`~kivy.properties.OptionProperty` and defaults to `off`.
    '''

    flash_frequency = NumericProperty(1)
    '''The number of times per second the display should flash.

    :attr:`flash_frequency` is an :class:`~kivy.properties.NumericProperty` and defaults to 1.
    '''

    flash_mask = StringProperty(None)
    '''Contains the flash mask string to use when flashing in mask mode. Each character of the flash
    mask string represents a character in the display. Character positions with an `F` character
    will be flashed while any other character will not flash.

    :attr:`flash_mask` is an :class:`kivy.properties.StringProperty` and defaults to None.
    '''


widget_classes = [SegmentDisplayEmulator]



import os

__version__ = '0.57.0.dev12'
__short_version__ = '0.57'
__bcp_version__ = '1.1'
__config_version__ = '6'
__mpf_version_required__ = '0.57.0.dev32'  # ALSO UPDATE in pyproject.toml

# pylint: disable-msg=invalid-name
version = f"MPF-MC v{__version__}"
'''A friendly version string for this build of MPF.'''

# pylint: disable-msg=invalid-name
extended_version = f"MPF-MC v{__version__} (config_version={__config_version__}, BCP v{__bcp_version__}, \
                     Requires MPF API compatibility v{ __mpf_version_required__} or newer)"



from kivy.properties import DictProperty
from kivy.properties import NumericProperty
from kivy.uix.effectwidget import EffectBase

class LinearGradientEffect(EffectBase):
    """GLSL effect to apply a linear gradient to a texture."""

    color_stops = DictProperty()
    '''This defines the colors of the gradient at each point.

    The key of this dictionary is the position [0, 1] along the gradient and the value is the
    RGBA color at that position.
    '''

    angle = NumericProperty(0.0)
    '''This defines the angle of the gradient.'''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.do_glsl()

    def on_color_stops(self, *args):
        self.do_glsl()

    def on_angle(self, *args):
        self.do_glsl()

    def do_glsl(self):
        linear_gradient_stops_glsl = self.__create_gradient_stops_glsl()
        self.glsl = linear_gradient_glsl.format(linear_gradient_stops_glsl, float(self.angle))

    def __create_gradient_stops_glsl(self):
        sorted_positions = sorted(self.color_stops)
        positions = len(sorted_positions)
        
        if not positions:
            return ""
        elif positions == 1:
            first_color = next(iter(self.color_stops.values()))
            return "gradient_color = {};".format(self.__rgba_list_to_vec4_glsl(first_color))
        
        stops = list()
        for current_stop_index in range(1, len(sorted_positions)):
            previous_stop_index = current_stop_index - 1
            
            previous_position = sorted_positions[previous_stop_index]
            current_position = sorted_positions[current_stop_index]
            
            previous_color = self.__rgba_list_to_vec4_glsl(self.color_stops[previous_position]) 
            current_color = self.__rgba_list_to_vec4_glsl(self.color_stops[current_position])
            
            stop_glsl = self.__create_gradient_stop_glsl(
                    "gradient_color" if current_stop_index > 1 else previous_color,
                    current_color,
                    previous_position,
                    current_position)
            
            stops.append(stop_glsl)
        
        return "\n".join(stops)

    def __create_gradient_stop_glsl(self, old_color, new_color, old_position, new_position):
        return "gradient_color = mix({}, {}, smoothstep({}, {}, position));".format(
                old_color,
                new_color,
                float(old_position),
                float(new_position))

    def __rgba_list_to_vec4_glsl(self, channels):
        red = float(channels[0])
        green = float(channels[1])
        blue = float(channels[2])
        alpha = float(channels[3])
        
        return "vec4({}, {}, {}, {})".format(red, green, blue, alpha)


linear_gradient_glsl = '''
        vec4 linear_gradient_color(float position)
        {{
            vec4 gradient_color = vec4(1.0);
            {}
            return gradient_color;
        }}
        
        vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords)
        {{
            // To apply the gradient at an angle, we convert the UV coordinates to polar, rotate
            // the point by the desired angle, then convert back to cartesian and use either the
            // x or y axis as the gradient position. A linear gradient changes color along one axis
            // only so it doesn't matter which axis we use to sample the gradient; however, it will
            // impact what the default angle of the gradient is when rendered.
            
            // Unlike most high-level languages, GLSL trig functions operate in radians not degrees
            float gradient_angle = radians({});
            
            // To convert the UV to polar, we must first shift it so the origin is in the center,
            // otherwise we would be constrained to one quarter of the coordinate space since UVs
            // are never negative.
            tex_coords -= 0.5;
            
            // Next, we convert the coordinates to polar
            float theta = atan(tex_coords.y, tex_coords.x);
            float distance = length(tex_coords);
            
            // Now we apply the desired angle offset, convert back to cartesian, move the
            // origin back to the lower-left to make this a valid UV, and finally sample the
            // x-coordinate to use it as the gradient position.
            // 
            // Since we're only sampling the x-coordinate, we don't need worry about computing the
            // y-coordinate in cartesian. We choose to sample x here instead of y because sampling
            // x will result in a horizonal gradient with the 0 position on the left. A horizontal
            // gradient with 0 on the left matches how gradients are represented in most image
            // editing software so this may be more intuitive as the default.
            float gradient_position = cos(theta + gradient_angle) * distance + 0.5;
            
            vec4 gradient_color = linear_gradient_color(gradient_position);
            
            // Perform a 'multiply' blend mode. Other blend modes could be added later but multiply
            // is simple and applicable to many use-cases.
            return color * gradient_color;
        }}
        '''

effect_cls = LinearGradientEffect
name = 'linear_gradient'



from mpfmc.tests.MpfIntegrationTestCase import MpfIntegrationTestCase
from mpfmc.tests.MpfSlideTestCase import MpfSlideTestCase
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestSegmentDisplayEmulator(MpfIntegrationTestCase, MpfFakeGameTestCase, MpfSlideTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'integration/machine_files/segment_display_emulator/'

    def test_segment_display_widget(self):
        """Integration test to test segment_display_player in MPF updating segment display widget in MC"""
        self.advance_time_and_run()
        self.assertTextOnTopSlide("")

        self.post_event("update_segment_display_1")
        self.advance_time_and_run()
        self.assertTextOnTopSlide("  HELLO")

        # this event updates a different segment display (not on top slide) -- text should not change
        self.post_event("update_segment_display_2")
        self.advance_time_and_run()
        self.assertTextOnTopSlide("  HELLO")

        self.post_event("update_segment_display_5")
        self.advance_time_and_run()
        self.assertTextOnTopSlide(" SCROLL")

        self.post_event("update_segment_display_3")
        self.advance_time_and_run()
        self.assertTextOnTopSlide("* BYE *")



"""Contains the parent class for custom code used in MPF-MC."""


class CustomCode:

    """Baseclass for custom code in a machine."""

    __slots__ = ["machine", "name", "delay"]

    def __init__(self, mc, name):
        """initialize custom code."""
        self.mc = mc
        self.name = name

        self.on_load()

    def __repr__(self):
        """Return string representation."""
        return '<CustomCode.{}>'.format(self.name)

    def on_load(self):
        """Automatically called when this custom code class loads.

        It's the intention that the programmer will overwrite this method
        in his custom code.
        """
        pass


from mpf.core.segment_mappings import SEVEN_SEGMENTS, FOURTEEN_SEGMENTS

from mpfmc.tests.MpfMcTestCase import MpfMcTestCase
from mpfmc.widgets.segment_display_emulator import SegmentDisplayEmulator


class TestSegmentDisplayEmulatorWidget(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/segment_display_widget'

    def get_config_file(self):
        return 'test_segment_display_widget.yaml'

    def test_character_encoding(self):
        self.assertEqual(2304,
                         SegmentDisplayEmulator.get_fourteen_segment_character_encoding(FOURTEEN_SEGMENTS[ord(')')]))
        self.assertEqual(11,
                         SegmentDisplayEmulator.get_seven_segment_character_encoding(SEVEN_SEGMENTS[ord(')')]))

    def test_segment_display(self):
        self.mc.events.post('show_top_display')
        self.mc.events.post('show_middle_display')
        self.mc.events.post('show_bottom_display')
        self.advance_real_time(1)
        self.mc.events.post('update_segment_display', segment_display_name='display1', text='GOODBYE')
        self.mc.events.post('update_segment_display', segment_display_name='display2', text='FOR NOW ')
        self.advance_real_time(0.033)
        self.assertIn('GOODBYE', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.assertIn('FOR NOW ', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.mc.events.post('update_segment_display', segment_display_name='display2', text='\x11               ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2', text='\x12\x11              ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text='\x13\x12\x11             ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text='\x13\x13\x12\x11            ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text='\x14\x13\x13\x12\x11           ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text=' \x14\x13\x13\x12\x11          ')
        self.advance_real_time(0.033)
        self.assertIn(' \x14\x13\x13\x12\x11          ', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text='W  \x14\x13\x13\x12\x11         ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text='OW \x14\x13\x13\x12\x11        ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text='NOW \x14\x13\x13\x12\x11       ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text=' NOW \x14\x13\x13\x12\x11      ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text='R NOW \x14\x13\x13\x12\x11     ')
        self.advance_real_time(0.033)
        self.assertIn('R NOW \x14\x13\x13\x12\x11     ', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text='OR NOW \x14\x13\x13\x12\x11    ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text='FOR NOW \x14\x13\x13\x12\x11   ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text=' FOR NOW \x14\x13\x13\x12\x11  ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text='  FOR NOW \x14\x13\x13\x12\x11 ')
        self.advance_real_time(0.033)
        self.mc.events.post('update_segment_display', segment_display_name='display2',
                            text='   FOR NOW \x14\x13\x13\x12\x11')
        self.advance_real_time(1)
        self.assertIn('   FOR NOW \x14\x13\x13\x12\x11', [x.widget.text for x in self.mc.targets[
            'default'].current_slide.widgets])



"""Mission Pinball Framework Media Controller (mpf-mc) setup.py.

Notes:
    See pyproject.toml for the rest of the setup config.
"""

# TODO: this setup has only been tested with the .c files already generated by cython. No idea if it works with the .pyx files yet.

import sys

if sys.platform != 'win32':
    import pkgconfig as pc

from collections import defaultdict
from os.path import abspath, exists, join, isdir
from setuptools import Extension, setup

sound_file_source = 'mpfmc/core/audio/sound_file.pyx',
track_source = 'mpfmc/core/audio/track.pyx',
track_standard_source = 'mpfmc/core/audio/track_standard.pyx',
track_sound_loop_source = 'mpfmc/core/audio/track_sound_loop.pyx',
audio_interface_source = 'mpfmc/core/audio/audio_interface.pyx',
playlist_controller_source = 'mpfmc/core/audio/playlist_controller.pyx',
bitmap_font_source = 'mpfmc/uix/bitmap_font/bitmap_font.pyx',

def members_appended(*ds):  # shoutout to the borg backup project
        result = defaultdict(list)
        for d in ds:
            for k, v in d.items():
                assert isinstance(v, list)
                result[k].extend(v)
        return result

def get_isolated_env_paths():  # shoutout to the kivy project
    try:
        # sdl2_dev is installed before setup.py is run, when installing from
        # source due to pyproject.toml. However, it is installed to a
        # pip isolated env, which we need to add to compiler
        import kivy_deps.sdl2_dev as sdl2_dev
    except ImportError:
        return [], []

    root = abspath(join(sdl2_dev.__path__[0], '../../../..'))
    includes = [join(root, 'Include')] if isdir(join(root, 'Include')) else []
    libs = [join(root, 'libs')] if isdir(join(root, 'libs')) else []
    return includes, libs

def determine_sdl2():  # shoutout to the kivy project
    flags = {}
    includes, _ = get_isolated_env_paths()

    flags['libraries'] = ['SDL2', 'SDL2_mixer', 'SDL2_image']

    sdl2_paths = []
    for include in includes + [join(sys.prefix, 'include')]:
        sdl_inc = join(include, 'SDL2')
        if isdir(sdl_inc):
            sdl2_paths.append(sdl_inc)

    flags['include_dirs'] = sdl2_paths
    flags['extra_link_args'] = []
    flags['extra_compile_args'] = []
    flags['library_dirs'] = sdl2_paths

    # ensure headers for all the SDL2 and sub libraries are available
    libs_to_check = ['SDL2', 'SDL2_mixer', 'SDL2_image']
    for lib in libs_to_check:
        found = False
        for d in flags['include_dirs']:
            fn = join(d, '{}.h'.format(lib))
            if exists(fn):
                found = True
                print('SDL2: found {} header at {}'.format(lib, fn))
                break

        if not found:
            print('SDL2: missing sub library {}'.format(lib))

    return flags

if sys.platform == 'win32':

    # Since Windows doesn't have pkgconfig, we have to manually find the SDL2 include location. :(

    sdl2_include_path = determine_sdl2()['include_dirs'][0]
    print("Setting SDL2 include path:", sdl2_include_path)

    general_include_path = sdl2_include_path.strip('\SDL2') # strip the SDL2 folder
    print("Setting general include path:", general_include_path)

    libs_include_path = general_include_path[:-8] # strip the \include
    libs_include_path = join(libs_include_path, 'libs')
    print("Setting libs include path:", libs_include_path)

    audio_kws = {'define_macros': [('_THREAD_SAFE', None)],
                 'include_dirs': [sdl2_include_path, general_include_path],
                'libraries': ['SDL2_mixer', 'SDL2', 'gstreamer-1.0', 'glib-2.0', 'gobject-2.0'],
                'library_dirs': [libs_include_path]}

    bitmap_font_kws = {'define_macros': [('_THREAD_SAFE', None)],
                 'include_dirs': [sdl2_include_path, general_include_path],
                 'libraries': ['SDL2', 'SDL2_image'],
                'library_dirs': [libs_include_path]}

else:
    audio_kws = members_appended(pc.parse('SDL2_mixer'), pc.parse('gstreamer-1.0'))

    try:
        bitmap_font_kws = members_appended(pc.parse('sdl2'), pc.parse('SDL2_image'))
    except pc.pkgconfig.PackageNotFoundError:
        # SDL2_image is not found sometimes on MacOS, not sure why, but it works with the mixer paths too
        bitmap_font_kws = members_appended(pc.parse('sdl2'), pc.parse('SDL2_mixer'))

ext_modules = [
    Extension('mpfmc.core.audio.sound_file', [*sound_file_source], **audio_kws),
    Extension('mpfmc.core.audio.track', [*track_source], **audio_kws),
    Extension('mpfmc.core.audio.track_standard', [*track_standard_source], **audio_kws),
    Extension('mpfmc.core.audio.track_sound_loop', [*track_sound_loop_source], **audio_kws),
    Extension('mpfmc.core.audio.audio_interface', [*audio_interface_source], **audio_kws),
    Extension('mpfmc.core.audio.playlist_controller', [*playlist_controller_source], **audio_kws),
    Extension('mpfmc.uix.bitmap_font.bitmap_font', [*bitmap_font_source], **bitmap_font_kws,)
]

setup(
    ext_modules=ext_modules,
)


"""Text Input widget.

The :class:`MpfTextInput` widget is used to all the player (or an operator) to
enter text. It can be linked to a :class:`Text` widget which displays the text
that's been entered so far.

"""
from collections import deque
from typing import Optional
from kivy.clock import Clock

from mpfmc.widgets.text import Text

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc     # pylint: disable-msg=cyclic-import,unused-import


class MpfTextInput(Text):

    """Text input widget."""

    widget_type_name = 'text_input'
    animation_properties = list()

    def __init__(self, mc: "MpfMc", config: dict, key: Optional[str] = None, **kwargs) -> None:
        """initialize text input.

        Note that this class is called *MpfTextInput* instead of *TextInput*
        because Kivy has a class called *TextInput* which collides with this
        one if they have the same name.
        """
        super().__init__(mc=mc, config=config, key=key)

        del kwargs

        self.linked_text_widget = None
        self.linked_text_widget_right_edge = None
        self.active = False
        self.registered_event_handlers = set()

        # setup the char list and set the current displayed char
        self.char_list = deque()
        self.char_list.extend(self.config['char_list'])
        self.char_list.append('back')
        self.char_list.append('end')
        self.current_list = self.char_list
        self.final_list = deque()
        self.final_list.append('back')
        self.final_list.append('end')
        self._is_blocking = False

        Clock.schedule_once(self.find_linked_text_widget)

    def __repr__(self) -> str:
        try:
            return '<TextInput Widget key={}>'.format(self.key)
        except AttributeError:
            return '<TextInput Widget>'

    def find_linked_text_widget(self, dt) -> None:
        del dt

        for target in self.mc.targets.values():
            w = target.find_widgets_by_key(self.config['key'])
            if w:
                if w[0] != self and isinstance(w[0], Text):
                    self.linked_text_widget = w[0]
                elif len(w) > 1 and isinstance(w[1], Text):
                    self.linked_text_widget = w[1]
                break

        if self.linked_text_widget:
            self.active = True
            self._register_events()

            if self.config['dynamic_x']:
                # bind() is weak meth, so we don't have to unbind
                self.linked_text_widget.bind(size=self.set_relative_position)
                self.linked_text_widget_right_edge = (
                    self.linked_text_widget.get_text_width() + self.linked_text_widget.x)

        self.jump(self.config['initial_char'])

    def _register_events(self) -> None:
        if not self.config['shift_left_event']:
            self.config['shift_left_event'] = (
                'text_input_{}_shift_left'.format(self.key))

        if not self.config['shift_right_event']:
            self.config['shift_right_event'] = (
                'text_input_{}_shift_right'.format(self.key))

        if not self.config['select_event']:
            self.config['select_event'] = (
                'text_input_{}_select'.format(self.key))

        if not self.config['abort_event']:
            self.config['abort_event'] = (
                'text_input_{}_abort'.format(self.key))

        if not self.config['force_complete_event']:
            self.config['force_complete_event'] = (
                'text_input_{}_force_complete'.format(self.key))

        for event in self.config.get('block_events', []):
            self.registered_event_handlers.add(self.mc.events.add_handler(
                event, self.block_enable, priority=2  # Priority avoids conflicts with select/abort event handlers
            ))
            self.mc.bcp_processor.register_trigger(event)
        for event in self.config.get('release_events', []):
            self.registered_event_handlers.add(self.mc.events.add_handler(
                event, self.block_disable, priority=2
            ))
            self.mc.bcp_processor.register_trigger(event)

        self.registered_event_handlers.add(self.mc.events.add_handler(
            self.config['shift_left_event'], self.shift, places=-1))
        self.mc.bcp_processor.register_trigger(self.config['shift_left_event'])
        self.registered_event_handlers.add(self.mc.events.add_handler(
            self.config['shift_right_event'], self.shift, places=1))
        self.mc.bcp_processor.register_trigger(self.config['shift_right_event'])
        self.registered_event_handlers.add(self.mc.events.add_handler(
            self.config['select_event'], self.select))
        self.mc.bcp_processor.register_trigger(self.config['select_event'])
        self.registered_event_handlers.add(self.mc.events.add_handler(
            self.config['abort_event'], self.abort))
        self.mc.bcp_processor.register_trigger(self.config['abort_event'])
        self.registered_event_handlers.add(self.mc.events.add_handler(
            self.config['force_complete_event'], self.complete))
        self.mc.bcp_processor.register_trigger(self.config['force_complete_event'])

    def _deregister_events(self) -> None:
        self.mc.events.remove_handlers_by_keys(self.registered_event_handlers)

    def jump(self, char: str) -> None:
        # Unfortunately deque.index() is Python 3.5+, so we have to do it this
        # way.

        if not self.current_list:
            return

        counts_remaining = len(self.current_list)
        while self.current_list[0] != char:
            self.current_list.rotate(1)
            counts_remaining -= 1

            if not counts_remaining:
                raise ValueError("Cannot set text_input character to "
                                 "{} as that is not a valid entry in the "
                                 "widget's character list".format(char))

        self.shift(0, True)

    def shift(self, places: int = 1, force: bool = False, **kwargs) -> None:
        del kwargs
        if self._is_blocking:
            return
        if self.active or force:
            self.current_list.rotate(-places)

            self.mc.post_mc_native_event('text_input_{}_active_character'.format(self.key),
                                         text=self.current_list[0])

            if self.current_list[0] == 'end':
                self.font_size = self.config['font_size'] / 2
                self.update_text('END')
            elif self.current_list[0] == ' ':
                self.font_size = self.config['font_size'] / 2
                self.update_text('SPACE')
            elif self.current_list[0] == 'back':
                self.font_size = self.config['font_size'] / 2
                self.update_text('BACK')
            else:
                self.font_size = self.config['font_size']
                self.update_text(self.current_list[0])

            if self.config['dynamic_x'] and self.linked_text_widget:
                self.set_relative_position()

    def select(self, **kwargs) -> None:
        del kwargs
        if not self.active:
            return

        if self.current_list[0] == 'back':

            if self.linked_text_widget.text:
                # set the text_entry text to whatever the last char was
                self.current_list = self.char_list
                self.jump(self.linked_text_widget.text[-1:])
                # remove the last char from the associated widget
                self.linked_text_widget.update_text(
                    self.linked_text_widget.text[:-1])

        elif self.current_list[0] == 'end':
            self.complete()

        else:
            self.linked_text_widget.update_text(
                self.linked_text_widget.text + self.current_list[0])

            if len(self.linked_text_widget.text) >= self.config['max_chars']:
                # only show back and end
                self.current_list = self.final_list
                self.jump("end")

            if len(self.linked_text_widget.text) > self.config['max_chars']:
                # we are done
                self.complete()

        self.mc.post_mc_native_event('text_input_{}_select'.format(self.key),
                                     text=self.linked_text_widget.text, length=len(self.linked_text_widget.text))

    def set_relative_position(self, *args) -> None:
        del args

        new_right_edge = (self.linked_text_widget.get_text_width() +
                          self.linked_text_widget.x)

        self.x += new_right_edge - self.linked_text_widget_right_edge
        self.linked_text_widget_right_edge = new_right_edge

        if self.x < (self.linked_text_widget_right_edge +
                     self.config['dynamic_x_pad']):
            self.x += (self.linked_text_widget_right_edge +
                       self.config['dynamic_x_pad'] - self.x)

    def complete(self, **kwargs) -> None:
        del kwargs
        self.done()
        self.mc.post_mc_native_event('text_input_{}_complete'.format(self.key),
                                     text=self.linked_text_widget.text)
        """event: text_input_(key)_complete

        desc: This event is posted by a *text_input* display widget when the
            entered text is finalized.

        args:
            text: A string of the final characters that were entered.
        """

    def abort(self, **kwargs) -> None:
        del kwargs
        self.done()
        self.mc.post_mc_native_event('text_input_{}_abort'.format(self.key),
                                     text=self.linked_text_widget.text)
        """event: text_input_(key)_abort

        desc: This event is posted by a *text_input* display widget when the
            entering process was aborted.

        args:
            text: A string of the characters that were entered so far.
        """

    def done(self) -> None:
        self._deregister_events()
        self.active = False
        self.text = ''
        self._container.parent.remove_widget(self)

    def prepare_for_removal(self) -> None:
        self.done()
        super().prepare_for_removal()

    def block_enable(self, **kwargs) -> None:
        del kwargs
        self._is_blocking = True

    def block_disable(self, **kwargs) -> None:
        del kwargs
        self._is_blocking = False


widget_classes = [MpfTextInput]
